@@ -2,11 +2,12 @@ from __future__ import annotations
 
 import asyncio
 import logging
+from dataclasses import dataclass
 from datetime import timedelta
 from enum import Enum, unique
-from typing import Any, Optional, Set, TypeVar, Union
+from typing import Any, Optional, Set, Union
 
-from .const import CMD_DELAY_NEXT_PACKET
+from .const import CMD_DELAY_NEXT_PACKET, PACKET_DELAY_CLEAR_TIME_DEFAULT
 from .packets import (
     BIN32_ABS,
     BIN32_NET,
@@ -21,13 +22,34 @@ LOG = logging.getLogger(__name__)
 
 PACKET_HEADER = bytes.fromhex("feff")
 API_RESPONSE_WAIT_TIME = timedelta(seconds=3)  # Time to wait for an API response
-PACKET_DELAY_CLEAR_TIME = timedelta(
-    seconds=3
-)  # Time to wait after a packet delay request so that GEM can finish sending any pending packets
 
 
-T = TypeVar("T")
-R = TypeVar("R")
+@dataclass(frozen=True)
+class PacketProtocolMessage:
+    """Base class for messages sent by a PacketProtocol."""
+
+    protocol: PacketProtocol
+
+
+@dataclass(frozen=True)
+class ConnectionMadeMessage(PacketProtocolMessage):
+    """Message sent when a new connection has been made to a protocol. Sent once shortly after creation of the protocol instance."""
+
+    pass
+
+
+@dataclass(frozen=True)
+class PacketReceivedMessage(PacketProtocolMessage):
+    """Message sent when a packet has been received by the protocol."""
+
+    packet: Packet
+
+
+@dataclass(frozen=True)
+class ConnectionLostMessage(PacketProtocolMessage):
+    """Message sent when a protocol loses its connection. exc is the exception that caused the connection to drop, if any."""
+
+    exc: Optional[BaseException]
 
 
 class PacketProtocol(asyncio.Protocol):
@@ -35,7 +57,7 @@ class PacketProtocol(asyncio.Protocol):
 
     def __init__(
         self,
-        queue: asyncio.Queue[Packet],
+        queue: asyncio.Queue[PacketProtocolMessage],
     ):
         """
         Create a new protocol instance.
@@ -47,25 +69,37 @@ class PacketProtocol(asyncio.Protocol):
         self._transport: Optional[asyncio.BaseTransport] = None
 
     def connection_made(self, transport: asyncio.BaseTransport) -> None:
+        LOG.info("%d: Connection opened", id(self))
+        assert self._transport is None
         self._transport = transport
+        self._queue.put_nowait(ConnectionMadeMessage(protocol=self))
 
-    def connection_lost(self, exc: Optional[Any]) -> None:
+    def connection_lost(self, exc: Optional[BaseException]) -> None:
         if exc is not None:
-            LOG.warning("Connection lost: {}".format(exc))
+            LOG.warning("%d: Connection lost due to exception", id(self), exc_info=exc)
         else:
-            LOG.info("Connection closed")
+            LOG.info("%d: Connection closed", id(self))
         self._transport = None
+        self._queue.put_nowait(ConnectionLostMessage(protocol=self, exc=exc))
 
     def data_received(self, data: bytes) -> None:
-        LOG.debug("Received {} bytes".format(len(data)))
+        LOG.debug("%d: Received %d bytes", id(self), len(data))
         self._buffer.extend(data)
         try:
             packet = self._get_packet()
             while packet is not None:
-                self._queue.put_nowait(packet)
+                self._queue.put_nowait(
+                    PacketReceivedMessage(protocol=self, packet=packet)
+                )
                 packet = self._get_packet()
-        except Exception as e:
-            LOG.exception("Exception while attempting to parse a packet.", e)
+        except Exception:
+            LOG.exception("%d: Exception while attempting to parse a packet.", id(self))
+
+    def close(self) -> None:
+        """Closes the underlying transport, if any."""
+        if self._transport:
+            self._transport.close()
+        self._transport = None
 
     def _get_packet(self) -> Optional[Packet]:
         """
@@ -75,7 +109,10 @@ class PacketProtocol(asyncio.Protocol):
 
             def skip_malformed_packet(msg: str, *args: Any, **kwargs: Any):
                 LOG.debug(
-                    "Skipping malformed packet due to " + msg + ". Buffer contents: %s",
+                    "%d Skipping malformed packet due to "
+                    + msg
+                    + ". Buffer contents: %s",
+                    id(self),
                     *args,
                     self._buffer,
                 )
@@ -83,7 +120,11 @@ class PacketProtocol(asyncio.Protocol):
 
             header_index = self._buffer.find(PACKET_HEADER)
             if header_index == -1:
-                LOG.debug("No header found. Discarding junk data: %s", self._buffer)
+                LOG.debug(
+                    "%d: No header found. Discarding junk data: %s",
+                    id(self),
+                    self._buffer,
+                )
                 self._buffer.clear()
                 continue
             del self._buffer[0:header_index]
@@ -91,9 +132,10 @@ class PacketProtocol(asyncio.Protocol):
             if len(self._buffer) < len(PACKET_HEADER) + 1:
                 # Not enough length yet
                 LOG.debug(
-                    "Not enough data in buffer yet ({} bytes): {}".format(
-                        len(self._buffer), self._buffer
-                    )
+                    "%d: Not enough data in buffer yet (%d bytes): %s",
+                    id(self),
+                    len(self._buffer),
+                    self._buffer,
                 )
                 return None
 
@@ -113,7 +155,9 @@ class PacketProtocol(asyncio.Protocol):
             if len(self._buffer) < packet_format.size:
                 # Not enough length yet
                 LOG.debug(
-                    "Not enough data in buffer yet ({} bytes)".format(len(self._buffer))
+                    "%d: Not enough data in buffer yet (%d bytes)",
+                    id(self),
+                    len(self._buffer),
                 )
                 return None
 
@@ -129,15 +173,17 @@ class PacketProtocol(asyncio.Protocol):
                     if len(self._buffer) < BIN48_NET_TIME.size:
                         # Not enough length yet
                         LOG.debug(
-                            "Not enough data in buffer yet ({} bytes)".format(
-                                len(self._buffer)
-                            )
+                            "%d: Not enough data in buffer yet (%d bytes)",
+                            id(self),
+                            len(self._buffer),
                         )
                         return None
 
                     result = BIN48_NET_TIME.parse(self._buffer)
 
-                LOG.debug("Parsed one {} packet.".format(result.packet_format.name))
+                LOG.debug(
+                    "%d: Parsed one %s packet.", id(self), result.packet_format.name
+                )
                 del self._buffer[0 : result.packet_format.size]
                 return result
             except MalformedPacketException as e:
@@ -186,13 +232,35 @@ class ProtocolStateException(Exception):
 class BidirectionalProtocol(PacketProtocol):
     """Protocol implementation for bi-directional communication with a GreenEye Monitor."""
 
-    def __init__(self, queue: asyncio.Queue[Packet]):
+    """
+    Create a new BidirectionalProtocol
+
+    The passed in queue contains full packets that have been received.
+    The packet_delay_clear_time plus API_RESPONSE_WAIT_TIME must be less than 15 seconds.
+    """
+
+    def __init__(
+        self,
+        queue: asyncio.Queue[PacketProtocolMessage],
+        packet_delay_clear_time: timedelta = PACKET_DELAY_CLEAR_TIME_DEFAULT,
+    ):
+        # Ensure that the clear time and the response wait time fit within the 15 second packet delay interval that is requested.
+        assert (packet_delay_clear_time + API_RESPONSE_WAIT_TIME) < timedelta(
+            seconds=15
+        )
+
         super().__init__(queue)
-        self._state = ProtocolState.RECEIVING_PACKETS
         self._api_buffer = bytearray()
+        self._packet_delay_clear_time = packet_delay_clear_time
+        self._state = ProtocolState.RECEIVING_PACKETS
+
+    @property
+    def packet_delay_clear_time(self) -> timedelta:
+        return self._packet_delay_clear_time
 
     def data_received(self, data: bytes) -> None:
         if self._state == ProtocolState.SENT_API_REQUEST:
+            LOG.debug("%d: Received %d bytes", id(self), len(data))
             self._api_buffer.extend(data)
         else:
             super().data_received(data)
@@ -207,13 +275,13 @@ class BidirectionalProtocol(PacketProtocol):
         """
         self._expect_state(ProtocolState.RECEIVING_PACKETS)
 
-        LOG.debug("Starting API request. Requesting packet delay...")
+        LOG.debug("%d: Starting API request. Requesting packet delay...", id(self))
         self._ensure_write_transport().write(
             CMD_DELAY_NEXT_PACKET.encode()
         )  # Delay packets for 15 seconds
         self._state = ProtocolState.SENT_PACKET_DELAY_REQUEST
 
-        return PACKET_DELAY_CLEAR_TIME
+        return self._packet_delay_clear_time
 
     def send_api_request(self, request: str) -> timedelta:
         """
@@ -225,7 +293,7 @@ class BidirectionalProtocol(PacketProtocol):
         """
         self._expect_state(ProtocolState.SENT_PACKET_DELAY_REQUEST)
 
-        LOG.debug(f"Sending API request '{request}'...")
+        LOG.debug("%d: Sending API request '%s'...", id(self), request)
         self._ensure_write_transport().write(request.encode())
         self._state = ProtocolState.SENT_API_REQUEST
 
@@ -240,7 +308,7 @@ class BidirectionalProtocol(PacketProtocol):
         self._expect_state(ProtocolState.SENT_API_REQUEST)
 
         response = bytes(self._api_buffer).decode()
-        LOG.debug(f"Received API response: '{response}'")
+        LOG.debug("%d: Received API response: '%s'", id(self), response)
         self._state = ProtocolState.RECEIVED_API_RESPONSE
 
         return response
@@ -257,7 +325,7 @@ class BidirectionalProtocol(PacketProtocol):
             }
         )
         self._api_buffer.clear()
-        LOG.debug(f"Ended API request")
+        LOG.debug("%d: Ended API request", id(self))
         self._state = ProtocolState.RECEIVING_PACKETS
 
     def _ensure_write_transport(self) -> asyncio.WriteTransport:
